#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <linux/vt.h>
#include <linux/fb.h>

#include <signal.h>

#include "ff.h"
#include "fb.h"

fftw_real Avg[N/2+1], AvgCount;

/* If zero, must not interfere the screen. */
/* If was zero and became nonzero,   *
 * must refresh the screen contents. */
int Active=1;

/* The current tty handle */
static int tty;

int COL =640;
int LINE=480;

unsigned char *Vid;
unsigned char Palette[256*3];

int FontHeight=0;
unsigned char *FontData;

/* The initial value in this is not much        *
 * meaningful. Just let it be 42 and it's fine. */
double FPS=42;

/* Framebuffer filehandle */
int fb_fh;

static Viewer *Vw;

static void SetSwitcher(void);
static void Switcher(int signal)
{
	if(signal==SIGUSR1)
	{
		ioctl(tty, VT_RELDISP, 1);
		Active=0;
	}
	if(signal==SIGUSR2)
	{
		ioctl(tty, VT_RELDISP, VT_ACKACQ);
		Active=1;
	}
	
	SetSwitcher();
}

static void SetSwitcher(void)
{
	signal(SIGUSR1, Switcher);
	signal(SIGUSR2, Switcher);
}

static void Setp(int pos, int bits, int c)
{
    register int i;
    register unsigned char *ptr = Vid+pos;

    for (i = 7; i >= 0; i--)
        *(ptr++) = (bits & (1 << i)) ? c : 0;
}
static void Putc(int pos, int c)
{
    int i;
    c *= FontHeight;
    for(i = 0; i < FontHeight; i++)
	{
        Setp(pos, FontData[c++], 136 + i*60 / FontHeight);
        pos += COL;
    }
}
void fb_center(int y, char *s)
{
	int pos = COL*y + (COL - strlen(s) * 8) / 2;
	
	while(*s)
	{
		Putc(pos, *s++);
		pos += 8;
	}
}	

void fb_init(Viewer *viewer)
{
	struct fb_var_screeninfo vscr;
	struct vt_mode           vt_mode;
	
	FILE *fp;
	
    if(isatty(0))
    	tty = 0;
    else 
		Trap(tty = open("/dev/tty",O_RDWR));
    	
    SetSwitcher();

	Vw = viewer;
	
	if(!Vw)return;
	
	if(Vw->fb_required == fb_is_required)
	{
	    Trap(ioctl(tty,VT_GETMODE, &vt_mode));
    
	    vt_mode.mode   = VT_PROCESS;
	    vt_mode.waitv  = 0;
	    vt_mode.relsig = SIGUSR1;
	    vt_mode.acqsig = SIGUSR2;

	    Trap(ioctl(tty,VT_SETMODE, &vt_mode));
		Trap(fb_fh=open("/dev/fb0", O_WRONLY));
		Trap(ioctl(fb_fh, FBIOGET_VSCREENINFO, &vscr));
	
		COL  = vscr.xres_virtual;
		LINE = vscr.yres_virtual;

		Trap((Vid = (unsigned char *)malloc(COL*LINE)) ? 1 : -1);	
	}
	else
	{
		COL = 80;
		LINE= 24;
	}
	
	FontHeight = 8;
	fp = fopen("cp850-8x8", "rb");

	FontData = (unsigned char *)malloc(FontHeight * 256);
	if(fp)
	{
		fread(FontData, 256, FontHeight, fp);
		fclose(fp);	
	}
	else
	{
		memset(FontData, 0, FontHeight*256);
		/* Or should we just fail here? */
	}
	
	bzero(Avg, sizeof Avg);
	AvgCount = 0.0;	

	Vw->Init();
}

void fb_view(time_t *Begin, time_t *Now)
{
	int a;
	
	#if 0
	AvgCount++;		
	for(a=0; a <= N/2; a++)Avg[a] += Power[a];
	
	/* Also [0] is calculated in Avg!! */
		
	/* Fix up (mix up) the average values twice in minute... */
	if(AvgCount > FPS*30)
	{
		AvgCount /= 2;
		*Begin = (*Begin + *Now) / 2;
		for(a=0; a <= N/2; a++)Avg[a] /= 2;
	}
	
	if(*Now != *Begin)
		FPS = AvgCount / (double)(*Now - *Begin);
	#endif
	
	if(Vw)
	{
		Vw->View();
		if(Active && Vw->fb_required==fb_is_required && !Vw->fb_refresh_self)
		{
			lseek(fb_fh, 0, SEEK_SET);
			write(fb_fh, Vid, LINE*COL);
		}
	}
}

void fb_done(void)
{
	if(!Vw)return;
	
	/* Reset terminal */
	
	Vw->Done();
	
	if(Vw->fb_required==fb_is_required)printf("\33[c");
	printf("\33[m");
	if(Vw->fb_required==fb_is_required)printf("\33[2J\33[H");
	printf("Spectrum analyzer terminated\n");
	
	if(Vw->fb_required==fb_is_required)
	{
		free(Vid);
		Vid = NULL;
	
		close(fb_fh);
	}
	Vw = NULL;	
}

void fb_writepalette(void)
{
	int a;
	struct fb_cmap cmap;	
	unsigned char *Pal = Palette + 3*16;
	
	if(!Vw)return;
	if(Vw->fb_required != fb_is_required)return;
	
	cmap.start=0;
	cmap.len  =256;
	cmap.red  =calloc(256, 2);
	cmap.green=calloc(256, 2);
	cmap.blue =calloc(256, 2);
	cmap.transp=NULL;	

	ioctl(fb_fh, FBIOGETCMAP, &cmap);
	for(a=16; a<256; a++)
	{
		cmap.red  [a] = *Pal++ * 65535 / 255;
		cmap.green[a] = *Pal++ * 65535 / 255;
		cmap.blue [a] = *Pal++ * 65535 / 255;
	}
	ioctl(fb_fh, FBIOPUTCMAP, &cmap);
	
	free(cmap.red);
	free(cmap.green);
	free(cmap.blue);
}
